/* ball.c
**
**    code for creating the GL-ball display-lists
**    Copyright (C) 2001  Florian Berger
**    Email: harpin_floh@yahoo.de, florian.berger@jk.uni-linz.ac.at
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License Version 2 as
**    published by the Free Software Foundation;
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/

#include <GL/glu.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "billard.h"
#include "billmove.h"
#include "ball.h"
#include "png_loader.h"
#include "options.h"


#define USE_VERTEX_ARRAYS
//#undef USE_VERTEX_ARRAYS


static char * balltexdata[20];
static GLuint balltexbind[20];
static int    balltexw,balltexh;
static int    shadowtexw,shadowtexh;
static char * shadowtexdata;
static GLuint shadowtexbind;
static int    depth;

static GLfloat col_spec [4] = {0.0, 0.0, 0.0, 1.0}; /* dont need any specular because of reflections */
static GLfloat col_spec2[4] = {1.0, 1.0, 1.0, 1.0}; /* dont need any specular because of reflections */
static GLfloat col_refl [4] = {1.0, 1.0, 1.0, 0.28};
static GLfloat col_diff [4] = {0.7, 0.7, 0.7, 1.0};
static GLfloat col_diff2[4] = {0.75, 0.75, 0.75, 1.0};
static GLfloat col_amb  [4] = {0.2, 0.2, 0.2, 1.0};
static GLfloat col_amb2 [4] = {0.45, 0.45, 0.45, 1.0}; /* for simple reflections */
static GLfloat col_shad [4] = {0.5, 0.0, 0.0, 0.0};


void normalize(float *v) /*FOLD00*/
{
    GLfloat d;
    d = sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
    v[0] /= d; v[1] /= d; v[2] /= d;
}


void rescale(float *v, float r)
{
    GLfloat d;
    d = sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2])/r;
    v[0] /= d; v[1] /= d; v[2] /= d;
}


/*
void normalize_my(myfloat *v)
{
    GLfloat d;
    d = sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
    v[0] /= d; v[1] /= d; v[2] /= d;
}
*/


static int in_array( ElemArray * array, GLfloat * v, GLfloat tex_x, GLfloat tex_y )
{
    int i;

    for(i=0;i<array->vnr;i++){
        if( v[0]==array->vert[i*3+0] &&
            v[1]==array->vert[i*3+1] &&
            v[2]==array->vert[i*3+2] &&
            tex_x==array->tex[i*2+0] &&
            tex_y==array->tex[i*2+1]
          ){
            return i;
        }
    }

    return -1;
}


void createvertex( float * v, float * n,
                   float tex_x, float tex_y, ElemArray * array )
{
    int pos;
    if ( array == NULL ){
//        fprintf(stderr,"creating glMultiTexCoord2fARB\n");
//        glMultiTexCoord2fARB(GL_TEXTURE0_ARB,tex_x, tex_y );
//        glMultiTexCoord2fARB(GL_TEXTURE1_ARB,tex_x, tex_y );
        glTexCoord2f( tex_x, tex_y );
        glNormal3fv(n);
        glVertex3fv(v);
    } else {
        pos = in_array( array,v, tex_x, tex_y );
        if( pos!=-1 ){
            array->index[array->indnr] = pos;
            array->indnr++;
        } else {
            array->vert[array->vnr*3+0]=v[0];
            array->vert[array->vnr*3+1]=v[1];
            array->vert[array->vnr*3+2]=v[2];
            array->norm[array->vnr*3+0]=n[0];
            array->norm[array->vnr*3+1]=n[1];
            array->norm[array->vnr*3+2]=n[2];
            array->tex[array->vnr*2+0]=tex_x;
            array->tex[array->vnr*2+1]=tex_y;
            array->index[array->indnr] = array->vnr;
            array->vnr++;
            array->indnr++;
            fprintf(stderr,"createvertex: array->vnr=%d\n",array->vnr);
        }
    }
}

// because used in ball_subdivide_rec
void ball_subdivide9_rec(float *v1, float *v2, float *v3, const int depth, float r, ElemArray *array);

void ball_subdivide_rec(float *v1, float *v2, float *v3, const int depth, float r, ElemArray *array)
/* if array=0 just make vertices */
{
    GLfloat v12[3], v23[3], v31[3], v123[3];
    GLfloat n1[3], n2[3], n3[3];
    GLint i;
    int otherside;
    double xt,yt,rho;

//    fprintf(stderr,"depth=%d\n",depth);
    if (depth == 0) {

        for(i=0;i<3;i++){
            n1[i]=v1[i];   n2[i]=v2[i];   n3[i]=v3[i];
        }
        normalize(n1);  normalize(n2);  normalize(n3);
        rescale(v1,r);  rescale(v2,r);  rescale(v3,r);

        otherside = ( v1[2]>0.0 || v2[2]>0.0 || v3[2]>0.0 );
//        glEnable(GL_NORMALIZE);
//        glTexCoord2f(0.5+0.5*v1[0], 0.5-0.5*(otherside?v1[1]:-v1[1]) );
        rho = sqrt(v1[0]*v1[0]+v1[1]*v1[1]);
        xt=asin(rho)/M_PI*2.3*v1[0]/rho;
        yt=asin(rho)/M_PI*2.3*v1[1]/rho;
        xt=v1[0]/r;
        yt=v1[1]/r;
        createvertex( v1, n1, 0.5+0.5*xt, 0.5-0.5*(otherside?yt:-yt), array);
//        glTexCoord2f(0.5+0.5*xt, 0.5-0.5*(otherside?yt:-yt) );
//        glNormal3fv(n1);
//        glVertex3fv(v1);

        rho = sqrt(v2[0]*v2[0]+v2[1]*v2[1]);
        xt=asin(rho)/M_PI*2.3*v2[0]/rho;
        yt=asin(rho)/M_PI*2.3*v2[1]/rho;
        xt=v2[0]/r;
        yt=v2[1]/r;
        createvertex( v2, n2, 0.5+0.5*xt, 0.5-0.5*(otherside?yt:-yt), array);
//        glTexCoord2f(0.5+0.5*xt, 0.5-0.5*(otherside?yt:-yt) );
//        glNormal3fv(n2);
//        glVertex3fv(v2);

        rho = sqrt(v3[0]*v3[0]+v3[1]*v3[1]);
        xt=asin(rho)/M_PI*2.3*v3[0]/rho;
        yt=asin(rho)/M_PI*2.3*v3[1]/rho;
        xt=v3[0]/r;
        yt=v3[1]/r;
        createvertex( v3, n3, 0.5+0.5*xt, 0.5-0.5*(otherside?yt:-yt), array);
//        glTexCoord2f(0.5+0.5*xt, 0.5-0.5*(otherside?yt:-yt) );
//        glNormal3fv(n3);
//        glVertex3fv(v3);

    } else if( depth==1 ){
        for (i = 0; i < 3; i++) v123[i] = v1[i]+v2[i]+v3[i];
        rescale(v123,r);
        ball_subdivide_rec( v123, v1, v2, depth-1, r, array);
        ball_subdivide_rec( v123, v2, v3, depth-1, r, array);
        ball_subdivide_rec( v123, v3, v1, depth-1, r, array);
    } else if( depth==3 ){
        ball_subdivide9_rec( v1, v2, v3, depth-2, r, array);
    } else {
        for (i = 0; i < 3; i++) {
            v12[i] = v1[i]+v2[i];
            v23[i] = v2[i]+v3[i];
            v31[i] = v3[i]+v1[i];
        }
        rescale(v12,r);
        rescale(v23,r);
        rescale(v31,r);
        ball_subdivide_rec( v1, v12, v31, depth-2, r, array);
        ball_subdivide_rec( v2, v23, v12, depth-2, r, array);
        ball_subdivide_rec( v3, v31, v23, depth-2, r, array);
        ball_subdivide_rec(v12, v23, v31, depth-2, r, array);
    }
//    fprintf(stderr,"back again all (depth=%d)\n",depth);
}


void ball_subdivide9_rec(float *v1, float *v2, float *v3, const int depth, float r, ElemArray *array)
{
    GLfloat v112[3], v122[3], v223[3], v233[3], v331[3], v311[3], v123[3];
    GLint i;

//    fprintf(stderr,"depth=%d\n",depth);
    if (depth == 0) {
        ball_subdivide_rec(v1,v2,v3,0,r, array); // draw the triangle
    } else {
        for (i = 0; i < 3; i++) {
            v112[i] = v1[i]+v1[i]+v2[i];
            v122[i] = v1[i]+v2[i]+v2[i];
            v223[i] = v2[i]+v2[i]+v3[i];
            v233[i] = v2[i]+v3[i]+v3[i];
            v331[i] = v3[i]+v3[i]+v1[i];
            v311[i] = v3[i]+v1[i]+v1[i];
            v123[i] = v1[i]+v2[i]+v3[i];
        }
        rescale(v112,r);
        rescale(v122,r);
        rescale(v223,r);
        rescale(v233,r);
        rescale(v331,r);
        rescale(v311,r);
        rescale(v123,r);
        ball_subdivide9_rec( v1  , v112, v311, depth-1, r, array);
        ball_subdivide9_rec( v112, v122, v123, depth-1, r, array);
        ball_subdivide9_rec( v112, v123, v311, depth-1, r, array);
        ball_subdivide9_rec( v311, v123, v331, depth-1, r, array);
        ball_subdivide9_rec( v122, v2  , v223, depth-1, r, array);
        ball_subdivide9_rec( v122, v223, v123, depth-1, r, array);
        ball_subdivide9_rec( v123, v223, v233, depth-1, r, array);
        ball_subdivide9_rec( v123, v233, v331, depth-1, r, array);
        ball_subdivide9_rec( v331, v233, v3  , depth-1, r, array);
    }
//    fprintf(stderr,"back again all (depth=%d)\n",depth);
}



void create_ball_icosa( float r, int ddepth, int id ) /*FOLD00*/
{
#define X .525731112119133606
#define Z .850650808352039932

    int i;

    static GLfloat vdata[12][3] = {
        {-X, 0.0, Z}, {X, 0.0, Z}, {-X, 0.0, -Z}, {X, 0.0, -Z},
        {0.0, Z, X}, {0.0, Z, -X}, {0.0, -Z, X}, {0.0, -Z, -X},
        {Z, X, 0.0}, {-Z, X, 0.0}, {Z, -X, 0.0}, {-Z, -X, 0.0}
    };

    static GLint tindices[20][3] = {
        {0,4,1}, {0,9,4}, {9,5,4}, {4,5,8}, {4,8,1},
        {8,10,1}, {8,3,10}, {5,3,8}, {5,2,3}, {2,7,3},
        {7,10,3}, {7,6,10}, {7,11,6}, {11,0,6}, {0,1,6},
        {6,1,10}, {9,0,11}, {9,11,2}, {9,2,5}, {7,2,11}
    };


    glDisable( GL_NORMALIZE );   /* remove this */
    glNewList(id, GL_COMPILE);
//      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
//      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
      glPushMatrix();
//      glScalef( r, r, r );
//      glEnable( GL_NORMALIZE );   /* remove this */
      glBegin(GL_TRIANGLES);
         for (i = 0; i < 20; i++) {
             ball_subdivide_rec(
                                &vdata[tindices[i][0]][0],
                                &vdata[tindices[i][1]][0],
                                &vdata[tindices[i][2]][0],
                                ddepth, r, NULL
                               );
         }
      glEnd();
      glPopMatrix();
//      glDisable( GL_NORMALIZE );   /* remove this */
    glEndList();

}



ElemArray * create_ball_icosa_array( float r, int ddepth, int id )
{
#define X .525731112119133606
#define Z .850650808352039932

    ElemArray * array;
    int i,pnr,fnr;
    double volume, scale_ratio;
    VMvect v1,v2,v3;

    static GLfloat vdata[12][3] = {
        {-X, 0.0, Z}, {X, 0.0, Z}, {-X, 0.0, -Z}, {X, 0.0, -Z},
        {0.0, Z, X}, {0.0, Z, -X}, {0.0, -Z, X}, {0.0, -Z, -X},
        {Z, X, 0.0}, {-Z, X, 0.0}, {Z, -X, 0.0}, {-Z, -X, 0.0}
    };

    static GLint tindices[20][3] = {
        {0,4,1}, {0,9,4}, {9,5,4}, {4,5,8}, {4,8,1},
        {8,10,1}, {8,3,10}, {5,3,8}, {5,2,3}, {2,7,3},
        {7,10,3}, {7,6,10}, {7,11,6}, {11,0,6}, {0,1,6},
        {6,1,10}, {9,0,11}, {9,11,2}, {9,2,5}, {7,2,11}
    };

    static int init = 0;


    if(!init){
        init=1;
         for (i = 0; i < 12; i++) {
             rescale(vdata[i],r);
         }
    }

    if(ddepth%2==0){
        fnr = 20*(1<<(ddepth/2*2));
        pnr = 10*((1<<(ddepth/2*2))-1) + 12;
    } else if(ddepth!=1) {
        fnr = 20*(1<<((ddepth-3)/2*2));
        pnr = 10*((1<<((ddepth-3)/2*2))-1) + 12;
        pnr += fnr * 4;
        fnr *= 9;
    } else if(ddepth==1) {
        fnr = 20*(1<<(ddepth/2*2));
        pnr = 10*((1<<(ddepth/2*2))-1) + 12;
        pnr += fnr; fnr *= 3;
    }
    pnr+=(int)sqrt(pnr)*2;  /* could get narrow */

    array = (ElemArray *)malloc( sizeof(ElemArray) );
    array->vert  = (GLfloat *)malloc( pnr*3*sizeof(GLfloat) );
    array->norm  = (GLfloat *)malloc( pnr*3*sizeof(GLfloat) );
    array->tex   = (GLfloat *)malloc( pnr*2*sizeof(GLfloat) );
    array->index = (int *)    malloc( fnr*3*sizeof(int)     );

    array->indnr=0;
    array->vnr=0;

    fprintf(stderr,"fnr*3=%d\n",fnr*3);
    fprintf(stderr,"pnr=%d\n",pnr);

    for (i = 0; i < 20; i++) {
        ball_subdivide_rec(
                           &vdata[tindices[i][0]][0],
                           &vdata[tindices[i][1]][0],
                           &vdata[tindices[i][2]][0],
                           ddepth, r, array
                          );
    }

    fprintf(stderr,"array->indnr=%d\n",array->indnr);
    fprintf(stderr,"array->vnr=%d\n",array->vnr);
/*
    glNewList(id, GL_COMPILE);
      glPushMatrix();
      glBegin(GL_TRIANGLES);
         for (i = 0; i < array->indnr; i++) {
             glTexCoord2fv( &array->tex [array->index[i]*2] );
             glNormal3fv(   &array->norm[array->index[i]*3] );
             glVertex3fv(   &array->vert[array->index[i]*3] );
             fprintf(stderr,"v=<%f,%f,%f>\n",
                     array->vert[array->index[i]*3],
                     array->vert[array->index[i]*3+1],
                     array->vert[array->index[i]*3+2] );
         }
      glEnd();
      glPopMatrix();
    glEndList();
*/

#if 1
    volume=0.0;
    for(i=0;i<array->indnr;i+=3){

        v1.x = array->vert[array->index[i+0]*3+0];
        v1.y = array->vert[array->index[i+0]*3+1];
        v1.z = array->vert[array->index[i+0]*3+2];

        v2.x = array->vert[array->index[i+1]*3+0];
        v2.y = array->vert[array->index[i+1]*3+1];
        v2.z = array->vert[array->index[i+1]*3+2];

        v3.x = array->vert[array->index[i+2]*3+0];
        v3.y = array->vert[array->index[i+2]*3+1];
        v3.z = array->vert[array->index[i+2]*3+2];

        volume+=tri_vol_xy( v1, v2, v3 );
    }
    scale_ratio=pow((4.0/3.0*r*r*r*M_PI)/fabs(volume),1.0/3.0);
    for(i=0;i<array->vnr*3;i++){
        array->vert[i] = array->vert[i]*scale_ratio;
    }
    glNewList(id, GL_COMPILE);
      glPushMatrix();
        /* call array */
        glEnableClientState( GL_VERTEX_ARRAY );
        glEnableClientState( GL_NORMAL_ARRAY );
        glEnableClientState( GL_TEXTURE_COORD_ARRAY );
        glVertexPointer( 3, GL_FLOAT, 0, array->vert );
        glNormalPointer( GL_FLOAT, 0, array->norm );
        glTexCoordPointer( 2, GL_FLOAT, 0, array->tex );

        fprintf(stderr,"indnr=%d\n",array->indnr);
        fprintf(stderr,"vnr=%d\n",array->vnr);

        glDrawElements( GL_TRIANGLES, array->indnr, GL_UNSIGNED_INT, array->index );

//      glEnd();
      glPopMatrix();
    glEndList();
#endif
    
//    fprintf(stderr,"array compiled %d, NE=%d, %d\n",glGetError(),GL_NO_ERROR, array);

    return array;
}



void draw_ball( BallType * ball, myvec cam_pos, GLfloat cam_FOV, int win_width ) /*FOLD00*/
/* expects to be on table center */
{
    float near, far, geomfact;
    static int glballlist[]={-1,-1,-1,-1,-1};
    static float ballmatr[]={
        0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0
    };
    int i,detail,maxdetail;
    static ElemArray * array[10];

    maxdetail=4;

    if( glballlist[0]==-1 ){
        for(i=0;i<=maxdetail;i++){
            glballlist[i] = glGenLists(1);
#ifdef USE_VERTEX_ARRAYS
            array[i] = create_ball_icosa_array( BALL_D/2.0, i, glballlist[i] );
#else
            create_ball_icosa( BALL_D/2.0, i, glballlist[i] );
#endif
        }
    }

    glPushMatrix();

    glTranslatef( ball->r.x, ball->r.y, ball->r.z );

    ballmatr[ 0]=ball->b[0].x;
    ballmatr[ 1]=ball->b[0].y;
    ballmatr[ 2]=ball->b[0].z;

    ballmatr[ 4]=ball->b[1].x;
    ballmatr[ 5]=ball->b[1].y;
    ballmatr[ 6]=ball->b[1].z;

    ballmatr[ 8]=ball->b[2].x;
    ballmatr[ 9]=ball->b[2].y;
    ballmatr[10]=ball->b[2].z;

    ballmatr[15]=1.0;

    glMultMatrixf( ballmatr );

/*    if(showReflections){
        glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
        glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
        glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
        glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
        glEnable(GL_TEXTURE_2D);
        glEnable(GL_TEXTURE_GEN_S);
        glEnable(GL_TEXTURE_GEN_T);
    }*/

//    glEnable(GL_TEXTURE_2D);

    geomfact=40.0/cam_FOV*(double)win_width/800;
    near=0.7*geomfact; far=5.5*geomfact;
    detail=(int)((double)maxdetail-(vec_abs(vec_diff(cam_pos,ball->r))-near)/(far-near)*(double)maxdetail);
    if( detail>maxdetail ) detail=maxdetail;
    if( detail<0 ) detail=0;
/*    #ifdef USE_VERTEX_ARRAYS
      glPushMatrix();
        glEnableClientState( GL_VERTEX_ARRAY );
        glEnableClientState( GL_NORMAL_ARRAY );
        glEnableClientState( GL_TEXTURE_COORD_ARRAY );
        glVertexPointer( 3, GL_FLOAT, 0, array[detail]->vert );
        glNormalPointer( GL_FLOAT, 0, array[detail]->norm );
        glTexCoordPointer( 2, GL_FLOAT, 0, array[detail]->tex );
        glDrawElements( GL_TRIANGLES, array[detail]->indnr, GL_UNSIGNED_INT, array[detail]->index );

      glPopMatrix();
    #else*/
      glCallList(glballlist[detail]);
/*    #endif*/

    glColor3f(1.0, 1.0, 1.0);
    glPopMatrix();

    /* ang mom vec */
/*    glPushMatrix();
    glTranslatef( ball->r.x, ball->r.y, ball->r.z );
      wabs=vec_abs(ball->w);
      vabs=vec_abs(ball->v);
      if( wabs ){
      glBegin(GL_LINES);
        glVertex3f( 0.0, 0.0, 0.0 );
        glVertex3f(
                    ball->w.x/wabs*0.1,
                    ball->w.y/wabs*0.1,
                    ball->w.z/wabs*0.1
                  );
        glVertex3f( 0.0, 0.0, 0.0 );
        glVertex3f(
                    ball->v.x/vabs*0.1,
                    ball->v.y/vabs*0.1,
                    ball->v.z/vabs*0.1
                  );
      glEnd();
      }
    glPopMatrix();*/
}


void draw_balls( BallsType balls, myvec cam_pos, GLfloat cam_FOV, int win_width, int spheretexbind )
{
    static int init = 0;
    int i;

    if( !init ){

        glGenTextures(1,&shadowtexbind);
//        load_png("shadow.png",&shadowtexw,&shadowtexh,&depth,&shadowtexdata);
//        load_png("shadow_gray_alpha.png",&shadowtexw,&shadowtexh,&depth,&shadowtexdata);
        load_png("shadow_alpha.png",&shadowtexw,&shadowtexh,&depth,&shadowtexdata);
        glBindTexture(GL_TEXTURE_2D,shadowtexbind);
//        gluBuild2DMipmaps(GL_TEXTURE_2D, 4, shadowtexw, shadowtexh, GL_RGBA,
//                          GL_UNSIGNED_BYTE, shadowtexdata);
//        gluBuild2DMipmaps(GL_TEXTURE_2D, 2, shadowtexw, shadowtexh, GL_LUMINANCE_ALPHA,
//                          GL_UNSIGNED_BYTE, shadowtexdata);
        gluBuild2DMipmaps(GL_TEXTURE_2D, 1, shadowtexw, shadowtexh, GL_ALPHA,
                          GL_UNSIGNED_BYTE, shadowtexdata);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, options_tex_min_filter);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, options_tex_mag_filter);
        fprintf(stderr,"shadowtexbind=%d\n",shadowtexbind);

        {
            char str[80];
            int i;
            for(i=0;i<16;i++){
                glGenTextures(1,&balltexbind[i]);
                sprintf(str,"billball%02d.png",i);
                load_png(str,&balltexw,&balltexh,&depth,&balltexdata[i]);
                glBindTexture(GL_TEXTURE_2D,balltexbind[i]);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, options_tex_min_filter);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, options_tex_mag_filter);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
                glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
                //            glTexImage2D(GL_TEXTURE_2D, 0, 3, balltexw, balltexh, 0,
                //                         GL_RGB, GL_UNSIGNED_BYTE, balltexdata[i]);
                gluBuild2DMipmaps(GL_TEXTURE_2D, 3, balltexw, balltexh, GL_RGB,
                                  GL_UNSIGNED_BYTE, balltexdata[i]);
                free(balltexdata[i]);
                fprintf(stderr,"balltexbind[%d]=%d\n",i,balltexbind[i]);
            }
        }

        init=1;
    }


/*    glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB,&i);
    fprintf(stderr,"max_ARB=%d\n",i);*/

    glEnable(GL_TEXTURE_2D);

    if( options_ball_reflect && !options_ball_reflections_blended ){
        glActiveTextureARB(GL_TEXTURE1_ARB);
        glBindTexture(GL_TEXTURE_2D,spheretexbind);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        glEnable(GL_TEXTURE_GEN_S);
        glEnable(GL_TEXTURE_GEN_T);
        glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
        glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
        glEnable(GL_TEXTURE_2D);
        glActiveTextureARB(GL_TEXTURE0_ARB);
    }
        //    glClientActiveTextureARB(GL_TEXTURE0_ARB);

    /* draw balls */
    if( options_ball_reflect && !options_ball_reflections_blended ){
        glMaterialfv(GL_FRONT, GL_DIFFUSE,   col_diff2);
        glMaterialfv(GL_FRONT, GL_AMBIENT,   col_amb2 );
        glMaterialfv(GL_FRONT, GL_SPECULAR,  col_spec2);
        glMaterialf (GL_FRONT, GL_SHININESS, 20.0     );
    }else{
        glMaterialfv(GL_FRONT, GL_DIFFUSE,   col_diff);
        glMaterialfv(GL_FRONT, GL_AMBIENT,   col_amb );
        glMaterialfv(GL_FRONT, GL_SPECULAR,  col_spec);
        glMaterialf (GL_FRONT, GL_SHININESS, 0.0     );
    }
    if( !options_ball_tex ){
        glDisable(GL_TEXTURE_2D);
    }
    if( options_ball_tex || (options_ball_reflect && !options_ball_reflections_blended) ){
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);
        for(i=0;i<balls.nr;i++) if(balls.ball[i].in_game){
            glBindTexture(GL_TEXTURE_2D,balltexbind[balls.ball[i].nr]);
            draw_ball(&balls.ball[i],cam_pos,cam_FOV,win_width);
        }
    }

    if( options_ball_reflect && !options_ball_reflections_blended ){
        glActiveTextureARB(GL_TEXTURE1_ARB);
        glDisable(GL_TEXTURE_2D);
        glActiveTextureARB(GL_TEXTURE0_ARB);
    }

    /* draw extra blended ball-reflections */
    if( options_ball_reflect && options_ball_reflections_blended ){
        glDepthMask (GL_FALSE);
        glEnable(GL_BLEND);
        glBlendFunc (GL_SRC_ALPHA, GL_ONE);
        glMaterialfv(GL_FRONT, GL_DIFFUSE, col_refl);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
        glBindTexture(GL_TEXTURE_2D,spheretexbind);
        glEnable(GL_TEXTURE_GEN_S);
        glEnable(GL_TEXTURE_GEN_T);
        glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
        glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
        for(i=0;i<balls.nr;i++) if(balls.ball[i].in_game){
            draw_ball(&balls.ball[i],cam_pos,cam_FOV,win_width);
        }
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);
        glDisable(GL_BLEND);
        glDepthMask (GL_TRUE);
    }

   /* draw shadows */
    if( options_ball_shadows ){
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);
        glMaterialfv(GL_FRONT, GL_DIFFUSE, col_shad);
        glDepthMask (GL_FALSE);
        glEnable(GL_BLEND);
//        glBlendFunc (GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
        glBlendFunc (GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
        // glBlendFunc ( GL_ONE, GL_SRC_ALPHA );
        // glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
        glBindTexture(GL_TEXTURE_2D,shadowtexbind);
#define SH_SZ 0.047
        for(i=0;i<balls.nr;i++) if(balls.ball[i].in_game){
            glBegin(GL_POLYGON);
            glNormal3f( 0.0,0.0,1.0 );
            glTexCoord2f(0.0,1.0);
            glVertex3f( balls.ball[i].r.x-SH_SZ, balls.ball[i].r.y+SH_SZ, balls.ball[i].r.z-balls.ball[i].d/2.01 );
            glTexCoord2f(1.0,1.0);
            glVertex3f( balls.ball[i].r.x+SH_SZ, balls.ball[i].r.y+SH_SZ, balls.ball[i].r.z-balls.ball[i].d/2.01 );
            glTexCoord2f(1.0,0.0);
            glVertex3f( balls.ball[i].r.x+SH_SZ, balls.ball[i].r.y-SH_SZ, balls.ball[i].r.z-balls.ball[i].d/2.01 );
            glTexCoord2f(0.0,0.0);
            glVertex3f( balls.ball[i].r.x-SH_SZ, balls.ball[i].r.y-SH_SZ, balls.ball[i].r.z-balls.ball[i].d/2.01 );
            glEnd();
        }
        glDisable(GL_BLEND);
        glDepthMask (GL_TRUE);
    }

}
